{-# LANGUAGE LambdaCase      #-}
{-# LANGUAGE RecordWildCards #-}

module SortImports where

import Control.Monad (void)
import Data.List (groupBy, sort)
import Data.Maybe (isJust)

import Text.Megaparsec
import Text.Megaparsec.Char

import SortImports.Types
  ( Constructors(..)
  , Export(..)
  , Exposure(..)
  , LineType(..)
  , Module(..)
  , Parser
  , View(..)
  , Qualified(..)
  )

rword :: String -> Parser String
rword s = string s <* space1

identifierSuffix :: Parser String
identifierSuffix = many $ alphaNumChar <|> oneOf "_'"

lowerIdentifier :: Parser String
lowerIdentifier = (:) <$> (lowerChar <|> char '_') <*> identifierSuffix

upperIdentifier :: Parser String
upperIdentifier = (:) <$> upperChar <*> identifierSuffix

moduleSuffix :: Parser String
moduleSuffix = (:) <$> char '.' <*> upperIdentifier

moduleName :: Parser String
moduleName = concat <$> ((:) <$> upperIdentifier <*> many moduleSuffix <* space)

alias :: Parser String
alias = rword "as" *> upperIdentifier <* space

operatorChar :: Parser Char
operatorChar = oneOf "!#$%&*+-./:<=>?@\\^|~"

constructors :: Parser Constructors
constructors = do
  void $ char '('
  cs <- [] <$ string ".." <|> ((upperIdentifier <|> operator') `sepBy1` (space *> char ',' *> space))
  void $ char ')'
  pure $ if null cs then All else Explicit cs

operator' :: Parser String
operator' = do
  void $ char '('
  n <- some operatorChar
  void $ char ')'
  pure $ "(" <> n <> ")"

type' :: Parser Export
type' = do
  typeName <- upperIdentifier <|> operator'
  void $ skipMany (char ' ')
  cs <- optional constructors
  pure $ Type typeName cs

operator :: Parser Export
operator = do
  void $ char '('
  s <- some operatorChar
  void $ char ')'
  pure $ Operator s

function :: Parser Export
function = Function <$> lowerIdentifier

export :: Parser Export
export = function <|> type' <|> operator

exports :: Parser [Export]
exports = do
  void $ char '('
  space
  es <- (space *> export <* space) `sepBy` char ','
  space
  void $ char ')'
  pure es

exposure :: Parser (Exposure [Export])
exposure = do
  hiding <- isJust <$> optional (rword "hiding")
  (if hiding then Hidden else Exposed) <$> exports

module' :: Parser Module
module' = do
  void $ rword "import"
  _qualified' <- isJust <$> optional (rword "qualified")
  _name      <- moduleName
  _qualified'' <- isJust <$> optional (rword "qualified")
  _alias     <- optional alias
  _exports   <- optional exposure
  space
  let _qualified = if _qualified' then Just Qualified_Pre else if _qualified'' then Just Qualified_Post else Nothing
  pure Module {..}

sortExports :: Module -> Module
sortExports m = m { _exports = fmap sort <$> _exports m }

lineType :: String -> LineType
lineType x =
  case parse module' "" x of
    Left  _ -> CodeLine x
    Right m -> ModuleLine m

groupLines :: [LineType] -> [[LineType]]
groupLines = groupBy f
  where f (CodeLine _) (CodeLine _) = True
        f (ModuleLine _) (ModuleLine _) = True
        f _ _ = False

sortIfModules :: [LineType] -> [LineType]
sortIfModules []                  = []
sortIfModules xs@(CodeLine _:_)   = xs
sortIfModules xs@(ModuleLine _:_) = sort . fmap f $ xs
  where f (ModuleLine x) = ModuleLine $ sortExports x
        f x = x

sortImports :: Maybe Qualified -> String -> String
sortImports mq
  = unlines
  . fmap (\case CodeLine x -> x
                ModuleLine x -> view $ convert x)
  . concatMap sortIfModules
  . groupLines
  . map lineType
  . lines
  where
    convert x = x { _qualified = mq <|> _qualified x}
